#include <stdio.h>
#include <windows.h>
#include "privilege.h"

void print_help() {
    printf("Usage:\n"
           "  wkill <options> [<params>]\n"
           "Options:\n"
           "  -c, --caption <Window Caption>  Caption of the window to be closed\n"
           "  -a, --all                       Kill all matching windows.\n"
           "                                    By default only the 1st found window is closed\n"
           "  -e, --exact                     Matching window captions exactly.\n"
           "                                    By default all captions beginning from <Window Caption>\n"
           "                                    are matched.\n"
           "  -i                              Use case insensetive comparison.\n"
           "  -p, --pid <Process ID>          Kill by process ID (PID).\n"
           "  -n, --process-name <exac name>  Kill by process (executable) name.\n"
           "  -x, --exit-code <Exit code>     Process exit code (-1 by default).\n"
           "Examples:\n"
           "  wkill -c \"Microsoft Word\"\n"
           "  wkill -p 440\n"
          );
    exit(0);
} // end print_help()

BOOLEAN killall = FALSE;
BOOLEAN exact   = FALSE;
BOOLEAN case_sens = TRUE;
int kill_count=0;

BOOL
CALLBACK
kill_by_caption_EnumWindowsProc(
    HWND hwnd,      // handle to parent window
    LPARAM lParam   // application-defined value
    )
{
    BOOL    bAttachTh;              // indicates can we attach the process
    DWORD   dwCurThreadId = 0;      // id of current process
    DWORD   dwBadWndThreadId = 0;   // bad window thread id   
    DWORD   dwBadWndProcId;         // process id of "Security Warning" window

    PWCHAR  WinCap = (PWCHAR)lParam;
    PWCHAR  WinCap2 = NULL;

    ULONG len, len2;

    BOOL res = TRUE;
    BOOLEAN match = FALSE;

    len2 = GetWindowTextLengthW(hwnd);
    if(!len2)
        return TRUE;
    len = wcslen(WinCap);
    WinCap2 = (PWCHAR)GlobalAlloc(GMEM_FIXED, (len2+1)*sizeof(WCHAR));
    if(!WinCap2)
        return TRUE;
    GetWindowTextW(hwnd, WinCap2, len2*sizeof(WCHAR));

    if(len2 < len) {
        len = len2;
    }
    if(!exact) {
        if(case_sens) {
            match = !wcsncmp(WinCap, WinCap2, len);
        } else {
            match = !wcsnicmp(WinCap, WinCap2, len);
        }
    } else {
        if(case_sens) {
            match = !wcscmp(WinCap, WinCap2);
        } else {
            match = !wcsicmp(WinCap, WinCap2);
        }
    }
    if(match) {

        // Get process ID's
        dwCurThreadId = GetCurrentThreadId();
        //dwBadWndThreadId = GetWindowThreadProcessId( hwnd, &dwBadWndProcId);

        //bAttachTh = AttachThreadInput( dwCurThreadId, dwBadWndThreadId, TRUE );

        //SetFocus(hwnd);
    //    if ((SetFocus(hwnd) != NULL) && (bAttachTh))
        {
            //SetActiveWindow( hwnd );
            PostMessage(hwnd, WM_CLOSE, 0, 0);
        }
        if(!killall) {
            res = FALSE;
        }
        kill_count++;
    }
    GlobalFree(WinCap2);

    return res;
} // end kill_by_caption_EnumWindowsProc()

int
kill_by_caption(
    PWCHAR WinCap
    )
{
    EnumWindows(kill_by_caption_EnumWindowsProc, (LONG)WinCap);
    return 0;
} // end kill_by_caption()

int
kill_by_pid(
    ULONG pid,
    ULONG ecode
    )
{
    HANDLE hProcess;

    hProcess=OpenProcess(PROCESS_TERMINATE, FALSE, pid);

    if (hProcess == NULL) {
        printf("Can't open specified process (PID=%d)\n", pid);
        return -1;
    }

    if(!TerminateProcess(hProcess, ecode)) {
        printf("Can't kill process (PID=%d)\n", pid);
        return -1;
    }
    printf("Process (PID=%d) killed\n", pid);
    kill_count++;

    return 0;
} // end kill_by_caption()

ULONG (__stdcall *NtQuerySystemInformation) (
    ULONG   SystemInformationClass,
    PVOID   SystemInformation,
    ULONG   SystemInformationLength,
    PULONG  ReturnLength
);

typedef LONG    NTSTATUS;
typedef LONG    KPRIORITY;

#define STATUS_INFO_LENGTH_MISMATCH      (0xC0000004L)

#define SystemProcessesAndThreadsInformation    5

typedef struct _CLIENT_ID {
    DWORD         UniqueProcess;
    DWORD         UniqueThread;
} CLIENT_ID;

typedef struct _UNICODE_STRING {
    USHORT        Length;
    USHORT        MaximumLength;
    PWSTR         Buffer;
} UNICODE_STRING;

typedef struct _VM_COUNTERS {
    SIZE_T        PeakVirtualSize;
    SIZE_T        VirtualSize;
    ULONG         PageFaultCount;
    SIZE_T        PeakWorkingSetSize;
    SIZE_T        WorkingSetSize;
    SIZE_T        QuotaPeakPagedPoolUsage;
    SIZE_T        QuotaPagedPoolUsage;
    SIZE_T        QuotaPeakNonPagedPoolUsage;
    SIZE_T        QuotaNonPagedPoolUsage;
    SIZE_T        PagefileUsage;
    SIZE_T        PeakPagefileUsage;
} VM_COUNTERS;

typedef struct _SYSTEM_THREADS {
    LARGE_INTEGER KernelTime;
    LARGE_INTEGER UserTime;
    LARGE_INTEGER CreateTime;
    ULONG         WaitTime;
    PVOID         StartAddress;
    CLIENT_ID     ClientId;
    KPRIORITY     Priority;
    KPRIORITY     BasePriority;
    ULONG         ContextSwitchCount;
    LONG          State;
    LONG          WaitReason;
} SYSTEM_THREADS, * PSYSTEM_THREADS;

typedef struct _SYSTEM_PROCESSES {
    ULONG             NextEntryDelta;
    ULONG             ThreadCount;
    ULONG             Reserved1[6];
    LARGE_INTEGER     CreateTime;
    LARGE_INTEGER     UserTime;
    LARGE_INTEGER     KernelTime;
    UNICODE_STRING    ProcessName;
    KPRIORITY         BasePriority;
    ULONG             ProcessId;
    ULONG             InheritedFromProcessId;
    ULONG             HandleCount;
    ULONG             Reserved2[2];
    VM_COUNTERS       VmCounters;
#if _WIN32_WINNT >= 0x500
    IO_COUNTERS       IoCounters;
#endif
    SYSTEM_THREADS    Threads[1];
} SYSTEM_PROCESSES, * PSYSTEM_PROCESSES;

DWORD
FindProcessPid(
    WCHAR *ImagePathName
    )
{
    ULONG Status;
    ULONG cbBuffer = 0x8000;
    PVOID pBuffer = NULL;
    ULONG l=0;
    ULONG ImagePathNameLen = wcslen(ImagePathName)*sizeof(WCHAR);
    ULONG pid;
    HMODULE hNtdll;

    hNtdll = GetModuleHandle("ntdll.dll");
    if(!hNtdll) {
        printf("NTDLL.DLL not loaded\n");
        return 0;
    }

    NtQuerySystemInformation = (unsigned long (__stdcall *)(ULONG   SystemInformationClass,
                                                            PVOID   SystemInformation,
                                                            ULONG   SystemInformationLength,
                                                            PULONG  ReturnLength)) GetProcAddress( hNtdll, "NtQuerySystemInformation" );

    do {
        pBuffer = GlobalAlloc(GMEM_FIXED, cbBuffer);

        Status = NtQuerySystemInformation(
                    SystemProcessesAndThreadsInformation,
                    pBuffer, cbBuffer, NULL);

        if (Status == STATUS_INFO_LENGTH_MISMATCH) {
            GlobalFree(pBuffer);
            cbBuffer *= 2;
        }
    } while (Status == STATUS_INFO_LENGTH_MISMATCH ||
             Status != 0);

    PSYSTEM_PROCESSES pProcesses = (PSYSTEM_PROCESSES)pBuffer;

    for (;;) {
        if ((pProcesses->ProcessName.Length == ImagePathNameLen) &&
            (memicmp(pProcesses->ProcessName.Buffer,ImagePathName, ImagePathNameLen) == 0)) {
            pid = pProcesses->ProcessId;
            GlobalFree(pBuffer);
            return pid;
        }
        if(!pProcesses->NextEntryDelta)
            break;
        l += pProcesses->NextEntryDelta;
        if(l > cbBuffer)
            break;
        pProcesses = (PSYSTEM_PROCESSES)(((LPBYTE)pProcesses)
                        + pProcesses->NextEntryDelta);
        if(l+pProcesses->NextEntryDelta > cbBuffer)
            break;
    }

    GlobalFree(pBuffer);
    return 0;

} // end FindProcessPid()


int
main (void)
{
    int argc;
    WCHAR** argv;
    WCHAR*  CmdLine;

    ULONG  pid = -1;
    ULONG  ecode = -1;
    WCHAR* WinCap = NULL;
    WCHAR* ProcName = NULL;

    int i;
    int rc = 1;

    printf("Process kill utility v0.2 (c) by Alter, 2005-2006\n"
           "Home site http://www.alter.org.ua\n");

    CmdLine = GetCommandLineW();
    argv = CommandLineToArgvW(CmdLine, &argc);

    if(argc < 3) {
        print_help();
    }

    for(i=1; i<argc; i++) {
        if(!argv[i] || argv[i][0] != '-') {
/*            if(!TargetFN) {
               TargetFN = argv[i];
            } else
            if(!LinkFN) {
               LinkFN = argv[i];
            } else {*/
                print_help();
                exit(0);
/*            }
            continue;*/
        }
        if(!wcscmp(argv[i], L"-h") ||
           !wcscmp(argv[i], L"--help")) {
            print_help();
            exit(0);
        } else
        if(!wcscmp(argv[i], L"-a") ||
           !wcscmp(argv[i], L"--all")) {
            killall = TRUE;
        } else
        if(!wcscmp(argv[i], L"-e") ||
           !wcscmp(argv[i], L"--exact")) {
            exact = TRUE;
        } else
        if(!wcscmp(argv[i], L"-i") /*||
           !wcscmp(argv[i], L"--exact")*/) {
            case_sens = FALSE;
        } else
        if(!wcscmp(argv[i], L"-c") ||
           !wcscmp(argv[i], L"--caption")) {

            i++;
            if(!argv[i] || i>=argc) {
                printf("Window caption not specified\n");
                print_help();
                exit(0);
            }
            WinCap = argv[i];
        } else
        if(!wcscmp(argv[i], L"-n") ||
           !wcscmp(argv[i], L"--process-name")) {

            i++;
            if(!argv[i] || i>=argc) {
                printf("Process name (executable) not specified\n");
                print_help();
                exit(0);
            }
            ProcName = argv[i];
        } else
        if(!wcscmp(argv[i], L"-p") ||
           !wcscmp(argv[i], L"--pid")) {
            i++;
            if(i>=argc) {
                printf("No Process ID\n");
                print_help();
                exit(0);
            }
            if(swscanf(argv[i], L"%d", &pid) != 1) {
                printf("Invalid Process ID: %S\n", argv[i]);
                exit(0);
            }
        } else
        if(!wcscmp(argv[i], L"-x") ||
           !wcscmp(argv[i], L"--exit-code")) {
            i++;
            if(i>=argc) {
                printf("No exit code\n");
                print_help();
                exit(0);
            }
            if(swscanf(argv[i], L"%d", &ecode) != 1) {
                printf("Invalid exit code: %S\n", argv[i]);
                exit(0);
            }
        } else
        {
            print_help();
            exit(0);
        }
    }

    GainAllPrivileges();

    if(WinCap) {
        kill_by_caption(WinCap);
    } else
    if(pid != -1) {
        kill_by_pid(pid, ecode);
    } else
    if(ProcName) {
        while(pid = FindProcessPid(ProcName)) {
            kill_by_pid(pid, ecode);
            if(!killall)
                break;
        }
        if(!kill_count) {
            printf("No matching process (%S) found\n", ProcName);
        }
    } else
    {
        printf("Target not specified\n");
        print_help();
        exit(0);
    }

    return rc;
} // end main()
